//---------------------------------------------------------------------------
#include <vcl.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#pragma hdrstop
#include "g12a.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"

/* WGS84 Coordinate Constants */
#define C          2.99792458e+8
#define mu         3.986005e+14
#define OMEGAeDOT  7.2921151467e-5
#define pi         3.1415926535898
#define L1freq     1575420000.0
#define F          -4.442807633e-10

#define WGS84a     6378137.0
#define WGS84b     6356752.314

typedef struct
{
    char    space;
    char    prnnum;
    short   fit;
    short   health;
    short   accuracy;
    float   IDOT;
    float   OMEGADOT;
    double  io;
    double  omega;
    double  OMEGAo;
    float   Cus;
    float   Cuc;
    float   Crs;
    float   Cis;
    float   Crc;
    float   Cic;
    long    toe;
    double  sqrtA;
    double  e;
    double  Mo;
    float   deltan;
    long    aode; /*Orbit data issue.*/
    float   af0;
    float   af1;
    float   af2;
    long    toc;
    long    aodc; /*Clock data issue.*/
    float   TGD;
    long    tow; /*Second of GPS week.*/
    short   wn;  /*GPS week number.*/
} Eph;

typedef struct
{
    unsigned char MPC34;
    long MPC33;
    long MPC32;
    double MPC31;
    double MPC30;
    unsigned char MPC29;
    unsigned char MPC28;
    char MPC27;
    unsigned char MPC26;
    unsigned char MPC25;
    long MPC24;
    long MPC23;
    double MPC22;
    double MPC21;
    unsigned char MPC20;
    unsigned char MPC19;
    char MPC18;
    unsigned char MPC17;
    unsigned char MPC16;
    long MPC15;
    long MPC14;
    double MPC13;
    double MPC12;
    unsigned char MPC11;
    unsigned char MPC10;
    char MPC9;
    unsigned char MPC8;
    unsigned char MPC7;
    unsigned char MPC6;
    unsigned char MPC5;
    unsigned char MPC4;
    unsigned char MPC3;
    unsigned char MPC2;
    unsigned short MPC1;
} MPC;

typedef struct
{
    unsigned short PDOP;
    float navtdot;
    float navzdot;
    float navydot;
    float navxdot;
    float navt;
    double navz;
    double navy;
    double navx;
    char sitename;
    long rcvtime;
} PBN;

typedef struct
{
    double  x;
    double  y;
    double  z;
} Pos;

typedef struct
{
    char    svprn;
    double  rcv_time;
    double  sat_time;
    double  prn;
    double  phase;
    Pos sat_pos;
} Raw;

Pos Satpos(Eph ephe,double);

TForm1 *Form1;

HANDLE hCom;
HANDLE hFile;

DWORD dwError;
DWORD NumWord;
DWORD dwEvtMask;

//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
    : TForm(Owner)
{
DCB dcb;
BOOL fSuccess;

hCom = CreateFile("COM3",GENERIC_READ|GENERIC_WRITE
                    ,0,NULL,OPEN_EXISTING,0,NULL);
hFile = CreateFile("g12a0.txt",GENERIC_WRITE
                    ,FILE_SHARE_READ,NULL,CREATE_ALWAYS,0,NULL);

if (hCom == INVALID_HANDLE_VALUE)
{
    dwError = GetLastError();
    Form1->Label1->Caption="COM3 Open Error";
}
else
    Form1->Label1->Caption="COM3 Open";

fSuccess = GetCommState(hCom, &dcb);

if (!fSuccess)
    Form1->Label2->Caption="GetCommState Error";
else
    Form1->Label2->Caption="GetCommState Success";

dcb.BaudRate = 9600;
dcb.ByteSize = 8;
dcb.Parity = NOPARITY;
dcb.StopBits = ONESTOPBIT;
dcb.fInX = FALSE;
dcb.fOutX = FALSE;
dcb.fOutxDsrFlow = TRUE;
dcb.fOutxCtsFlow = TRUE;
dcb.fDsrSensitivity = TRUE;
dcb.fDtrControl = DTR_CONTROL_HANDSHAKE;
dcb.fRtsControl = RTS_CONTROL_HANDSHAKE;

fSuccess = SetCommState(hCom, &dcb);

if (!fSuccess) {
    Form1->Label2->Caption="SetCommState Error";
}
else
    Form1->Label2->Caption="SetCommState Success";

SetCommMask(hCom,EV_RXCHAR|EV_BREAK|EV_CTS|EV_DSR|EV_ERR
                 |EV_RING|EV_RLSD|EV_TXEMPTY);

//PurgeComm(hCom,PURGE_RXCLEAR|PURGE_TXCLEAR	);
/*
WriteFile(hCom,"$PASHS,SPD,A,7\r\n",16,&NumWord,0);
while( !(dwEvtMask & EV_RXCHAR) )
      WaitCommEvent(hCom,&dwEvtMask,0);
*/

fSuccess = GetCommState(hCom, &dcb);
dcb.BaudRate = 19200;
fSuccess = SetCommState(hCom, &dcb);

PurgeComm(hCom,PURGE_RXCLEAR|PURGE_TXCLEAR);

}

__fastcall thread1::thread1(bool CreateSuspended)
    : TThread(CreateSuspended)
{
}

//---------------------------------------------------------------------------

void __fastcall thread1::Execute()
{
   char r;
   char head[6];
   COMMTIMEOUTS Timeouts;

   Eph ephe[37]={0};
   Raw rawdata[37]={0};
   char snv_eph_on[37]={0};
   char snv_chn_on[37]={0};

   char flag_mpc=0;
   char flag_pbn=0;
   char flag_do=0;

//   FILE *ephtest;

   unsigned int count=0,count1=0;

//   ephtest=fopen("asc0a.txt","w");

   GetCommTimeouts(hCom,&Timeouts);
   Timeouts.ReadIntervalTimeout=MAXDWORD;
   Timeouts.ReadTotalTimeoutMultiplier =0;
   Timeouts.ReadTotalTimeoutConstant =0;
   SetCommTimeouts(hCom,&Timeouts);

   PurgeComm(hCom,PURGE_RXCLEAR|PURGE_TXCLEAR);
   WriteFile(hCom,"$PASHS,RCI,1\r\n",14,&NumWord,0);
   while( !(dwEvtMask & EV_RXCHAR) )
      WaitCommEvent(hCom,&dwEvtMask,0);
   PurgeComm(hCom,PURGE_RXCLEAR|PURGE_TXCLEAR);
   WriteFile(hCom,"$PASHS,ELM,15\r\n",15,&NumWord,0);
   while( !(dwEvtMask & EV_RXCHAR) )
      WaitCommEvent(hCom,&dwEvtMask,0);
   PurgeComm(hCom,PURGE_RXCLEAR|PURGE_TXCLEAR);
   WriteFile(hCom,"$PASHS,OUT,A\r\n",14,&NumWord,0);
   while( !(dwEvtMask & EV_RXCHAR) )
      WaitCommEvent(hCom,&dwEvtMask,0);
   WriteFile(hCom,"$PASHS,OUT,A,SNV,PBN,MBN,BIN\r\n",30,&NumWord,0);

   for(;;)
   {
    WaitCommEvent(hCom,&dwEvtMask,0);
    if( dwEvtMask & EV_BREAK )
      {
      Form1->Label3->Caption="EV_BREAK";
      }
    if( dwEvtMask & EV_CTS )
      {
      Form1->Label3->Caption="EV_CTS";
      }
    if( dwEvtMask & EV_DSR )
      {
      Form1->Label3->Caption="EV_DSR";
      }
    if( dwEvtMask & EV_ERR )
      {
      }
    if( dwEvtMask & EV_RING )
      {
      Form1->Label3->Caption="EV_RING";
      }
    if( dwEvtMask & EV_RLSD )
      {
      }
    if( dwEvtMask & EV_RXCHAR )
      {
       char stop=0;
       do{
          ReadFile(hCom,&r,1,&NumWord,0);
          if(NumWord==1)
            {
             stop=1;
             head[5]=r;
             if((head[0]=='$')&&(head[1]=='P')&&(head[2]=='A')&&(head[3]=='S')&&
                   (head[4]=='H')&&(head[5]=='R'))
               {
                 char temp[5];

                  GetCommTimeouts(hCom,&Timeouts);
                  Timeouts.ReadIntervalTimeout=50;
                  Timeouts.ReadTotalTimeoutMultiplier =0;
                  Timeouts.ReadTotalTimeoutConstant =0;
                  SetCommTimeouts(hCom,&Timeouts);

                  ReadFile(hCom,&temp,5,&NumWord,0);

                  if( NumWord!=5)  Form1->Label2->Caption="error";

                  if((temp[1]=='S')&&(temp[2]=='N')&&(temp[3]=='V')
                          &&(temp[4]==','))
                     {
                       char snv[134],snvtemp[134];
                       Eph *ephe_PRN;

                       WriteFile(hFile,"SNV,",4,&NumWord,0);
                       ReadFile(hCom,&snv,134,&NumWord,0);

                       for(int i=0;i<130;i++)
                           snvtemp[129-i]=snv[i];
                       ephe_PRN=(Eph *)snvtemp;
                       ephe[ephe_PRN->prnnum+1]=*ephe_PRN;
                       ephe[ephe_PRN->prnnum+1].prnnum++;

                       snv_eph_on[ephe_PRN->prnnum+1]=1;

                       if((snv[132]!='\r')&&(snv[133]!='\n'))
                              Form1->Label2->Caption="error";

                       WriteFile(hFile,&snv,134,&NumWord,0);
                     }

                   if((temp[1]=='M')&&(temp[2]=='P')&&(temp[3]=='C')
                          &&(temp[4]==','))
                     {
                       char mpc[97],mpctemp[97],*pchar;
                       MPC *mpc_PRN;
                       double *pointprn,*pointphase;

                       WriteFile(hFile,"MPC,",4,&NumWord,0);
                       ReadFile(hCom,&mpc,97,&NumWord,0);

                       for(int i=0;i<95;i++)
                           mpctemp[94-i]=mpc[i];

                       pchar=mpctemp;

                       pointprn=(double *)(pchar+67);
                       pointphase=(double *)(pchar+75);

                       rawdata[(unsigned char)mpctemp[91]].prn=*pointprn;
                       rawdata[(unsigned char)mpctemp[91]].phase=*pointphase;

                       if(((unsigned char)mpctemp[94-7]==2)
                                   ||((unsigned char)mpctemp[94-7]==1))
                           snv_chn_on[(unsigned char)mpctemp[91]]=1;
                       else snv_chn_on[(unsigned char)mpctemp[91]]=0;

                       if((unsigned char)mpctemp[94-2]==0) flag_mpc=1;
                       else flag_mpc=0;

                       if((mpc[95]!='\r')&&(mpc[96]!='\n'))
                              Form1->Label2->Caption="error";

                       WriteFile(hFile,&mpc,97,&NumWord,0);
                      }

                   if((temp[1]=='P')&&(temp[2]=='B')&&(temp[3]=='N')
                          &&(temp[4]==','))
                     {
                       char pbn[58],pbntemp[58];
                       PBN *pbn_PRN;

                       WriteFile(hFile,"PBN,",4,&NumWord,0);
                       ReadFile(hCom,&pbn,58,&NumWord,0);

                       for(int i=0;i<56;i++)
                           pbntemp[55-i]=pbn[i];
                       pbn_PRN=(PBN *)pbntemp;

                       for(int i=0;i<37;i++)
                           {
                             rawdata[i].rcv_time=(pbn_PRN->rcvtime)/1000;
                             rawdata[i].sat_time= rawdata[i].rcv_time-rawdata[i].prn;
                           }
                       int satnum=0;
                       //if(flag_mpc==1)

                          for(int i=0;i<37;i++)
                             {
                              if((snv_chn_on[i]==1)&&(snv_eph_on[i]==1))
                                 {
                                   AnsiString time,date;
                                   time=TimeToStr(Time());
                                   date=DateToStr(Date());

                                   satnum++;
                                   rawdata[i].sat_pos.x=Satpos(ephe[i],rawdata[i].sat_time).x;
                                   rawdata[i].sat_pos.y=Satpos(ephe[i],rawdata[i].sat_time).y;
                                   rawdata[i].sat_pos.z=Satpos(ephe[i],rawdata[i].sat_time).z;

/*                                  fprintf(ephtest,"%s %s %d %.15f %.15f %.15f %.15f %.15f %.15f %.15f %.15f\n",
                                      &(date[1]),&(time[1]),
                                      i,rawdata[i].rcv_time,rawdata[i].sat_time,
                                      rawdata[i].prn,rawdata[i].phase,
                                      rawdata[i].sat_pos.x,
                                      rawdata[i].sat_pos.y,
                                      rawdata[i].sat_pos.z,
                                      pbn_PRN->navt);
*/                                 }
                             }
                          for(int i=0;i<37;i++) snv_chn_on[i]=0;
                          count++;

                       Form1->Label3->Caption=IntToStr(satnum);
                       if((pbn[56]!='\r')&&(pbn[57]!='\n'))
                              Form1->Label2->Caption="error";

                       WriteFile(hFile,&pbn,58,&NumWord,0);
/*
                          if(count!=floor((pbn_PRN->rcvtime)/1000/86400))
                             {
//                                fclose(ephtest);
                                CloseHandle(hFile);
                                count=floor((pbn_PRN->rcvtime)/1000/86400);
                                count1=count;
                                char *s,*no;
                                s=new char [12];
                                no=new char [5];
                                StrCopy(s,"g12a");
                                itoa(count1,no,10);
                                StrCat(s,no);
                                StrCat(s,".txt");
//                                ephtest=fopen(&s[1],"w");
                                hFile = CreateFile(s,GENERIC_WRITE
                                        ,FILE_SHARE_READ,NULL,CREATE_ALWAYS,0,NULL);
                                delete [] s;
                                delete [] no;
                             }
*/


                      }

                  Form1->Label1->Caption=0;
                  Form1->Label1->Caption=&temp[0];

                  GetCommTimeouts(hCom,&Timeouts);
                  Timeouts.ReadIntervalTimeout=MAXDWORD;
                  Timeouts.ReadTotalTimeoutMultiplier =0;
                  Timeouts.ReadTotalTimeoutConstant =0;
                  SetCommTimeouts(hCom,&Timeouts);
               }
             else
              {
               head[0]=head[1];
               head[1]=head[2];
               head[2]=head[3];
               head[3]=head[4];
               head[4]=head[5];
              }
            }
          else stop=0;
         }while(stop==1);

      }
    if( dwEvtMask & EV_TXEMPTY )
      {
       //Form1->Label3->Caption="EV_TXEMPTY";
      }
   }
}
//---------------------------------------------------------------------------
Pos Satpos(Eph ephe,double time)
{
    Pos satposition;
    double A, n0, n, t, tk, Mk, E, Ek, sinvk, cosvk, vk;
    double PHIk, deltauk, deltark, deltaik, uk, rk, ik;
    double deltatr, deltatsv;
    double xkprime, ykprime, OMEGAk;
    int i, j, k;

	t=time;

	A=pow(ephe.sqrtA,2);
	n0=sqrt(mu/pow(A,3));
	tk=t-ephe.toe;
	if (tk>302400.0) tk=tk-604800.0;
	if (tk<-302400.0) tk=tk+604800.0;
	n=n0+ephe.deltan*pi;
	Mk=ephe.Mo*pi+n*tk;

	E=Mk;
	for (j=0;j<20;j++)
	{
	    Ek=E+(Mk+ ephe.e*sin(E)-E)/(1- ephe.e *cos(E));
	    if (fabs(Ek-E)<1e-8)
		break;
	    E=Ek;
	}

	if ((t-ephe.toc)>302400.0) t=t-604800.0;
	if ((t-ephe.toc)<-302400.0) t=t+604800.0;
	deltatr=F*ephe.e*ephe.sqrtA*sin(Ek);
	deltatsv=ephe.af0+ephe.af1*(t-ephe.toc)+
		 ephe.af2*pow((t-ephe.toc),2)+
				 deltatr-ephe.TGD;
	t=t-deltatsv;
	tk=t-ephe.toe;
	if (tk>302400.0) tk=tk-604800.0;
	if (tk<-302400.0) tk=tk+604800.0;
	Mk=ephe.Mo*pi+n*tk;
	E=Mk;
	for (j=0;j<20;j++)
	{
	    Ek=E+(Mk+ephe.e*sin(E)-E)/(1-ephe.e*cos(E));
	    if (fabs(Ek-E)<1e-8)
		break;
	    E=Ek;
	}
	sinvk=sqrt(1-pow(ephe.e,2))*sin(Ek)/(1-ephe.e*cos(Ek));
	cosvk=(cos(Ek)-ephe.e)/(1-ephe.e*cos(Ek));
	vk=atan2(sinvk,cosvk);
	PHIk=vk+ephe.omega*pi;
	deltauk=ephe.Cus*sin(2*PHIk)+ephe.Cuc*cos(2*PHIk);
	deltark=ephe.Crc*cos(2*PHIk)+ephe.Crs*sin(2*PHIk);
	deltaik=ephe.Cic*cos(2*PHIk)+ephe.Cis*sin(2*PHIk);
	uk=PHIk+deltauk;
	rk=A*(1-ephe.e*cos(Ek))+deltark;
	ik=ephe.io*pi+deltaik+ephe.IDOT*pi*tk;
	xkprime=rk*cos(uk);
	ykprime=rk*sin(uk);
	OMEGAk=ephe.OMEGAo*pi+(ephe.OMEGADOT*pi-OMEGAeDOT)*tk-
	       OMEGAeDOT*ephe.toe;
	satposition.x=xkprime*cos(OMEGAk)-ykprime*cos(ik)*sin(OMEGAk);
	satposition.y=xkprime*sin(OMEGAk)+ykprime*cos(ik)*cos(OMEGAk);
	satposition.z=ykprime*sin(ik);
    return(satposition);

}
//---------------------------------------------------------------------------

void __fastcall TForm1::Button4Click(TObject *Sender)
{
 thread1 *threadm1 = new thread1(true);
 threadm1->Resume();

}
//---------------------------------------------------------------------------

